#! /usr/bin/python3

import DistUtilsExtra.auto

DistUtilsExtra.auto.setup(name='Foo',
      version='0.1',
      description="package with private Python extension",
      author='Maximilian Engelhardt',
      author_email='maxi@daemonizer.de',
      url='http://www.debian.org/',
)
